# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025 Greenbone AG

import unittest

from troubadix.helper.remove_comments import remove_comments


class RemoveCommentsTestCase(unittest.TestCase):
    def test_empty_string(self):
        input_content = ""
        expected_output = ""
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_no_comments(self):
        input_content = (
            "function detect_archlinux(sock, port, SCRIPT_DESC, is_pfsense) {"
        )
        expected_output = input_content
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_full_line_comments(self):
        input_content = "# comment\nfunction hello() {\n    # Another comment\n    return 42;"
        expected_output = "\nfunction hello() {\n\n    return 42;"
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_inline_comments(self):
        input_content = (
            "function hello(){ # A function\n    return 42; # The answer"
        )
        expected_output = "function hello(){ \n    return 42; "
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_hash_in_strings(self):
        input_content = "message = 'This # is not a comment';\nurl = \"http://example.com/#anchor\";"
        expected_output = input_content
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_mixed_quotes_and_comments(self):
        input_content = (
            "display('Hash: #'); # Real comment\n"
            's = "My string with # character"; # Another comment'
        )
        expected_output = (
            "display('Hash: #'); \ns = \"My string with # character\"; "
        )
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_complex_scenario(self):
        # ruff does not allow whitespace at end of lines so this looks a bit weird
        input_content = """function hello(){
    # Comment at start
    if (TRUE) {# inline comment
        display('Hello # not a comment');
    # Full comment line
        display("Hash: #");# end comment
    }
}"""

        expected_output = """function hello(){

    if (TRUE) {
        display('Hello # not a comment');

        display("Hash: #");
    }
}"""

        self.assertEqual(remove_comments(input_content), expected_output)

    def test_indented_comments(self):
        input_content = (
            "function func() {\n    # Indented comment\n    return 0;\n}"
        )
        expected_output = "function func() {\n\n    return 0;\n}"
        self.assertEqual(remove_comments(input_content), expected_output)

    def test_backslash_double_quotes(self):
        input_content = r'display("\"); # This is a comment'
        expected_output = r'display("\"); '
        self.assertEqual(remove_comments(input_content), expected_output)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from troubadix.plugin import LinterError, LinterWarning
from troubadix.plugins.using_display import CheckUsingDisplay

from . import PluginTestCase


class CheckUsingDisplayTestCase(PluginTestCase):
    def test_ok(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"cvss_base_vector", '
            'value:"AV:N/AC:L/Au:S/C:N/I:P/A:N");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        self.assertEqual(0, len(results))

    def test_using_display(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"cvss_base_vector", '
            'value:"AV:N/AC:L/Au:S/C:N/I:P/A:N");\n'
            '  display("FOO");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            'VT is using a display() without any if statement at line 3: display("FOO");',
            results[0].message,
        )

    def test_using_if_display(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"cvss_base_vector", '
            'value:"AV:N/AC:L/Au:S/C:N/I:P/A:N");\n'
            'if (0) display("FOO");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], LinterWarning)
        self.assertIn(
            "VT is using a display() inside an if statement but without debug check",
            results[0].message,
        )
        self.assertIn('if (0) display("FOO");', results[0].message)

    def test_using_comment_display(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"cvss_base_vector", '
            'value:"AV:N/AC:L/Au:S/C:N/I:P/A:N");\n'
            '# display("FOO");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        # Comments are removed, so display() won't be found
        self.assertEqual(0, len(results))

    def test_using_debug_if_display(self):
        """Test that display() inside a debug if statement is allowed"""
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            'if (debug) display("FOO");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        # Should be OK because it's in a debug if
        self.assertEqual(0, len(results))

    def test_display_in_string_ignored(self):
        path = Path("some/file.nasl")
        content = "str = 'display(\"FOO\")';\n"
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckUsingDisplay(fake_context)

        results = list(plugin.run())

        self.assertEqual(0, len(results))



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025 Greenbone AG

from troubadix.helper.text_utils import StringState


def remove_comments(file_content: str) -> str:
    """
    Remove all commented portions from file content while preserving string literals.
    This function:
    1. Removes content from lines that start with '#' (also after whitespace)
    2. For lines with inline comments, keeps only the content before the '#'
    3. Preserves '#' characters within string literals
    4. Maintains original line numbers by keeping empty lines
    5. Handles multiline strings
    Args:
        file_content: String containing the full file content
    Returns:
        String with comments removed
    """
    string_state = StringState()
    return "\n".join(
        [
            _remove_comments_in_line(line, string_state)
            for line in file_content.splitlines()
        ]
    )


def _remove_comments_in_line(line: str, state: StringState) -> str:
    if not state.in_string and line.lstrip().startswith("#"):
        return ""

    for i, char in enumerate(line):
        state.process_next_char(char)
        if char == "#" and not state.in_string:
            return line[:i]

    return line
